from pathlib import Path
from types import SimpleNamespace as sn

from . import data

def cli() -> sn:
    from . import env, language, settings
    cli = data.sns.from_dict(data.json.read(Path(__file__).parent.parent.parent / 'data/package_data.json'))
    cli.msgs = language.get_msgs(cli,
        language.generate_random_lang(excludes=['en']) if env.is_debug_mode() else language.get_sys_lang())
    settings.load(cli)
    return cli



import os, sys
from pathlib import Path
from types import SimpleNamespace as sn
from typing import List, Optional, Union
if sys.platform == 'win32' : import colorama ; colorama.init() # enable ANSI color support

from . import color as colors, data as datalib, pkg

try : terminal_width = os.get_terminal_size()[0]
except OSError : terminal_width = 80

current_ver = datalib.json.read(Path(__file__).parent.parent.parent / 'data/package_data.json')['version']
next_maj_ver = pkg.get_next_maj_ver(current_ver)
_warned_keys = { 'cli': set(), 'config': set() }

def data(msg: str, *args, no_newline: bool = False, **kwargs) -> None:
    print(f'\n{colors.bw}{msg.format(*args, **kwargs)}{colors.nc}', end='' if no_newline else None)
def dim(msg: str, *args, no_newline: bool = False, **kwargs) -> None:
    print(f'\n{colors.gry}{msg.format(*args, **kwargs)}{colors.nc}', end='' if no_newline else None)
def docs_url(cli: sn) -> None : tip(f'{cli.msgs.tip_FOR_MORE_HELP_VISIT}:\n{cli.urls.docs}')
def error(msg: str, *args, **kwargs) -> None : print(f'\n{colors.br}ERROR: {msg.format(*args, **kwargs)}{colors.nc}')
def help_cmd(cli: sn) -> None : info(f"{cli.msgs.log_TYPE} '{cli.cmds[0]} --help' {cli.msgs.log_FOR_AVAIL_OPTIONS}\n")
def info(msg: str, *args, end: str = '', **kwargs) -> None:
    print(f'\n{colors.by}{msg.format(*args, **kwargs)}{colors.nc}', end=end)
def line_break() : print()
def overwrite_print(msg: str, *args, **kwargs) -> None:
    sys.stdout.write('\r' + msg.format(*args, **kwargs).ljust(terminal_width)[:terminal_width])
def success(msg: str, *args, **kwargs) -> None : print(f'\n{colors.bg}{msg.format(*args, **kwargs)}{colors.nc}')
def tip(msg: str, *args, **kwargs) -> None : print(f'\n{colors.bc}TIP: {msg.format(*args, **kwargs)}{colors.nc}')
def version(cli: sn) -> None:
    print(f'\n{colors.by}{cli.name}\n{colors.bw}{cli.msgs.log_VERSION.lower()}: {cli.version}{colors.nc}')
def warn(msg: str, *args, **kwargs) -> None : print(f'\n{colors.bo}WARNING: {msg.format(*args, **kwargs)}{colors.nc}')

def cmd_docs_url_exit(cli: sn, msg: str = '', cmd: str = 'help') -> None:
    if msg : error(msg)
    help_cmd(cli)
    docs_url(cli)
    sys.exit(1)

def debug(msg: str, cli: Optional[sn] = None, *args, **kwargs) -> None:
    from . import env
    if not env.is_debug_mode() : return

    # Init --debug [target]
    debug_key=None
    debug_argidx = sys.argv.index('--debug') if '--debug' in sys.argv else sys.argv.index('-V')
    if debug_argidx +1 < len(sys.argv) and not sys.argv[debug_argidx +1].startswith('-'):
        debug_key = sys.argv[debug_argidx +1].replace('-', '_')

    if cli: # init data line
        if debug_key:
            data_val = getattr(cli.config, debug_key, f'cli.config key {debug_key!r} {cli.msgs.warn_NOT_FOUND.lower()}')
        else:
            data_val = cli.config
        msg += f'\n{colors.gry}{data_val}{colors.nc}'

    if args: # use 'em
        msg = msg.format(*args, **kwargs)

    print(f'\n{colors.by}DEBUG: {msg}{colors.nc}')

def package_vers(
    pkgs: List[str], results: Union[Optional[str], List[Optional[str]]], cli: sn, scheme: str = 'rainbow'
) -> None:
    results = [results] if not isinstance(results, list) else results
    scheme_colors = getattr(colors.schemes, scheme)
    for idx, (package, version) in enumerate(zip(pkgs, results)):
        color = scheme_colors[idx % len(scheme_colors)]
        msg = f'Python {version}' if version else f'{colors.gry}{cli.msgs.log_NO_PY_REQ_FOUND}{colors.nc}'
        info(f'{color}{package}:{colors.nc} {msg}')
    line_break()

def trunc(msg: str, end: str = '\n') -> None:
    truncated_lines = [
        line if len(line) < terminal_width else line[:terminal_width -4] + '...' for line in msg.splitlines()]
    print('\n'.join(truncated_lines), end=end)

def warn_legacy_option(cli: sn, flag: str, source: str) -> None:
    from . import settings
    warned_set = _warned_keys[source]
    if flag in warned_set : return
    canonical_key = settings.get_canonical_key(flag)
    msg = f"{ cli.msgs.warn_CONFIG_FILE_KEY if source == 'config' else cli.msgs.warn_CLI_OPTION } {flag!r}"
    if canonical_key:
        canonical_ctrl = getattr(settings.controls, canonical_key, None)
        if source == 'cli' and canonical_ctrl:
            flags = [arg for arg in getattr(canonical_ctrl, 'args', []) if arg.startswith('-')]
            if flag.startswith('-') and len(flag) == 2: # show short flag replacement
                display_key = min(flags, key=len) if flags else f"--{canonical_key.replace('_', '-')}"
            else: # show long flag replacement
                long_flags = [flag for flag in flags if flag.startswith('--')]
                display_key = long_flags[0] if long_flags else f"--{canonical_key.replace('_', '-')}"
        else:
            display_key = canonical_key
        msg += f' {cli.msgs.warn_HAS_BEEN_REPLACED_BY} {display_key!r}'
    else:
        msg += f' {cli.msgs.warn_NO_LONGER_HAS_ANY_EFFECT}'
    msg += f' {cli.msgs.warn_AND_WILL_BE_REMOVED} @ v{next_maj_ver}'
    warn(msg) ; warned_set.add(flag)



from pathlib import Path
from typing import Union

def atomic_write(file_path: Union[Path, str], data: str, encoding: str ='utf-8') -> None: # to prevent TOCTOU
    import os
    file_path = Path(file_path)
    file_path.parent.mkdir(parents=True, exist_ok=True)
    tmp_path = file_path.parent / f'.{file_path.name}.tmp'
    try:
        with open(tmp_path, 'w', encoding=encoding) as file:
            file.write(data) ; file.flush() ; os.fsync(file.fileno())
        os.replace(tmp_path, file_path) # atomic rename
    except Exception:
        if tmp_path.exists() : tmp_path.unlink()
        raise

def read(file_path: Union[Path, str], encoding: str = 'utf-8') -> str:
    with open(file_path, 'r', encoding=encoding) as file:
        return file.read()

def write(file_path: Union[Path, str], data: str, encoding: str = 'utf-8') -> None:
    with open(file_path, 'w', encoding=encoding) as file:
        file.write(data)



from pathlib import Path
import re, sys
from types import SimpleNamespace as sn
from typing import List, Optional

from . import data, log

def format_code(lang_code: str) -> str: # to match locale dir name (e.g., 'zh-tw' -> 'zh_TW')
    return re.sub(
        r'([a-z]{2,8})[-_]([a-z]{2})',
        lambda m: f'{m.group(1).lower()}_{m.group(2).upper()}',
        lang_code, flags=re.IGNORECASE
    )

def generate_random_lang(includes: Optional[List[str]] = None,
                         excludes: Optional[List[str]] = None) -> str:
    import random

    if includes is None : includes = []
    if excludes is None : excludes = []

    def get_locales() -> List[str]:

        # Read cache if found
        cache_dir = Path(__file__).parent.parent / '_cache'
        locale_cache = cache_dir / 'locales.json'
        if locale_cache.exists():
            try : return data.json.read(locale_cache)
            except Exception : pass

        # Discover pkg _locales
        locales_dir = Path(__file__).parent.parent.parent / 'data/_locales'
        if not locales_dir.exists() : return ['en']
        locales = []
        for entry in locales_dir.iterdir():
            if entry.is_dir() and re.match(r'^\w{2}[-_]?\w{0,2}$', entry.name):
                locales.append(entry.name)

        # Cache result
        cache_dir.mkdir(parents=True, exist_ok=True)
        data.json.write(locale_cache, locales)

        return locales

    locales = includes.copy() if includes else get_locales()

    # Filter out excludes
    exclude_set = set(excludes)
    locales = [locale for locale in locales if locale not in exclude_set]

    # Get random language
    random_lang = random.choice(locales) if locales else 'en'
    log.debug(f'Random language: {random_lang}')

    return random_lang

def get_msgs(cli: sn, lang_code: str = 'en') -> sn:
    from . import jsdelivr, url

    lang_code = format_code(lang_code)
    if getattr(get_msgs, 'cached', None) and lang_code == get_msgs.cached_lang:
        return get_msgs.cached # don't re-fetch same msgs

    msgs = data.json.flatten(data.json.read( # local ones
        Path(__file__).parent.parent.parent / 'data/_locales/en/messages.json'))

    if not lang_code.startswith('en'): # fetch non-English msgs from jsDelivr
        import is_unicode_supported, non_latin_locales
        if not hasattr(get_msgs, 'cached_unicode_support'):
            get_msgs.cached_unicode_support = is_unicode_supported() # type: ignore
        if lang_code.split('_')[0] in non_latin_locales and not get_msgs.cached_unicode_support: # type: ignore
            return sn(**msgs) # EN ones cuz non-Latin not supported
        msg_base_url = f'{jsdelivr.create_commit_url(cli, cli.commit_hashes.locales)}' \
                        '/src/get_min_py/data/_locales'
        msg_url = f'{msg_base_url}/{lang_code}/messages.json'
        for attempt in range(3):
            try: # fetch remote msgs
                msgs = data.json.flatten(data.json.read(url.get(msg_url)))
                break
            except Exception: # retry up to 2X (region-stripped + EN)
                if attempt == 2 : break
                msg_url = ( re.sub(r'([^_]*)_[^/]*(/.*)', r'\1\2', msg_url) # strip region before retrying
                    if attempt == 0 and '-' in lang_code else f'{msg_base_url}/en/messages.json') # else use EN msgs

    get_msgs.cached = msgs
    get_msgs.cached_lang = lang_code

    return sn(**msgs)

def get_sys_lang(cli: Optional[sn] = None) -> str:
    import os, subprocess
    try:
        if sys.platform == 'win32':
            return subprocess.run(
                ['powershell', '-Command', '(Get-Culture).TwoLetterISOLanguageName'],
                capture_output=True, text=True, check=True
            ).stdout.strip()
        else: # macOS/Linux
            for env_lang_var in ['LANG', 'LANGUAGE', 'LC_ALL', 'LC_MESSAGES', 'LC_NAME']:
                sys_lang = os.environ.get(env_lang_var)
                if sys_lang : return sys_lang.split('.')[0]
            return 'en'
    except Exception as err:
        if cli : log.error(f'{cli.msgs.err_FAILED_TO_FETCH_SYS_LANG}: {err}')
        return 'en'



nc = '\x1b[0m'
hex = {
    'br': '#ff0000',  'by': '#ffff00',  'bo': '#ffa500',   'bg': '#00ff00',
    'bw': '#ffffff',  'dg': '#008000', 'gry': '#808080',  'blk': '#000000', 'tlBG': '#008080'
}

def hex_to_ansi(hex_color: str) -> str:
    r = int(hex_color[1:3], 16)
    g = int(hex_color[3:5], 16)
    b = int(hex_color[5:7], 16)
    return f'\x1b[38;2;{r};{g};{b}m'

class _Schemes:
    @property
    def default(self) -> list[str]:
        return [hex_to_ansi(hex) for hex in [
            '#00e5bc', '#18c8ae', '#30ac9f', '#488f91', '#607383',
            '#775674', '#8f3966', '#a71d57', '#bf0049', '#9a1b5e'
        ]]
    @property
    def rainbow(self) -> list[str]:
        return [hex_to_ansi(hex) for hex in [
            '#e41a1c', '#ff7f00', '#ffff33', '#4daf4a', '#377eb8',
            '#984ea3', '#f781bf', '#999999', '#a65628', '#d95f02'
        ]]
schemes = _Schemes()

def __getattr__(hex_key: str) -> str: # add color.hex_key getters that return ANSI
    if hex_key in hex: return hex_to_ansi(hex[hex_key])
    raise AttributeError(f"module 'color' has no attribute '{hex_key}'")



from types import SimpleNamespace as sn
from typing import Any, Dict

def from_dict(obj: Dict[str, Any]) -> sn:
    for key, val in obj.items():
        if isinstance(val, dict):
            obj[key] = from_dict(val)
    return sn(**obj)



import json
from pathlib import Path
from typing import Any, Dict, Union

import json5

def flatten(json: Dict[str, Any], key: str = 'message') -> Dict[str, Any]: # eliminate need to ref nested keys
    flat_obj = {}
    for json_key in json:
        val = json[json_key]
        flat_obj[json_key] = val[key] if isinstance(val, dict) and key in val else val
    return flat_obj

def is_valid(file_path: Union[Path, str], format: str = 'json') -> bool:
    file_path = Path(file_path)
    if not file_path.exists():
        return False
    try : file_text = file_path.read_text(encoding='utf-8')
    except Exception:
        return False
    if format == 'json':
        try : json.loads(file_text) ; return True
        except Exception : return False
    elif format == 'json5':
        try : json5.loads(file_text) ; return True
        except Exception : return False
    else:
        raise ValueError(f"Unsupported format {format!r}. Expected 'json' or 'json5'")

def read(input: Union[Path, str], encoding: str = 'utf-8') -> Any:
    input_str = str(input)
    if input_str.endswith(('.json', '.json5')):
        with open(input_str, 'r', encoding=encoding) as file:
           return json5.load(file)
    else : return json5.loads(input_str)

def write(file_path: Union[Path, str], data: Any, encoding: str = 'utf-8', ensure_ascii: bool = False,
          style: str = 'pretty', atomic: bool = True, max_line_length: int = 120) -> None:
    from . import file
    from typing import Optional, List

    Path(file_path).parent.mkdir(parents=True, exist_ok=True)

    def format_compact(obj: Any, indent: int = 0, padded_key: Optional[str] = None) -> List[str]:
        indent_spaces = '  ' * indent
        line_prefix = padded_key if padded_key else indent_spaces

        if isinstance(obj, dict):

            # Try fit whole dict in 1 line
            kv_pairs = [f'"{key}": {json.dumps(val, separators=(",",":"), ensure_ascii=ensure_ascii)}'
                for key,val in obj.items()]
            single_line_dict = f'{line_prefix}{{ {", ".join(kv_pairs)} }}'
            if len(single_line_dict) <= max_line_length:
                return [single_line_dict]

            # Else split long line up
            lines = [line_prefix + '{']
            for idx, (key,val) in enumerate(obj.items()):
                inner_lines = format_compact(val, indent +1, f'  {indent_spaces}"{key}": ')
                for line in inner_lines : lines.append(line)
                if not idx == len(obj) -1 : lines[-1] += ',' # append comma except last line
            lines.append(indent_spaces + '}')
            return lines

        elif isinstance(obj, list):

            # Try fit whole list in 1 line
            single_line_list = line_prefix + json.dumps(obj, separators=(',', ':'), ensure_ascii=ensure_ascii)
            if len(single_line_list) <= max_line_length:
                return [single_line_list]

            # Else split long list up
            lines = [line_prefix + '[']
            if all(not isinstance(item, (dict, list)) for item in obj): # all items primitives, pack into lines
                list_items = [json.dumps(item, ensure_ascii=ensure_ascii) for item in obj]
                inner_indent = '  ' * (indent + 1)
                current_line_items = []
                for item in list_items:
                    candidate_line = ', '.join(current_line_items + [item]) if current_line_items else item
                    if len(inner_indent + candidate_line) + 1 <= max_line_length : current_line_items.append(item)
                    else: # current line full, flush/start new line
                        if current_line_items : lines.append(inner_indent + ', '.join(current_line_items) + ',')
                        current_line_items = [item]
                if current_line_items: # flush last line
                    lines.append(inner_indent + ', '.join(current_line_items))
            else: # mixed/complex items, format each recursively
                for idx, item in enumerate(obj):
                    inner_lines = format_compact(item, indent +1)
                    for line in inner_lines : lines.append(line)
                    if not idx == len(obj) -1 : lines[-1] += ','
            lines.append(indent_spaces + ']')
            return lines

        else: # primitive
            return [line_prefix + json.dumps(obj, ensure_ascii=ensure_ascii)]

    # Format JSON
    if style == 'pretty': # single key/val spans multi-lines
        json_str = json.dumps(data, indent=2, ensure_ascii=ensure_ascii)
    elif style == 'compact': # single key/val per line but honors max_line_length
        json_str = '\n'.join(format_compact(data))
    else: # minified to single line
        json_str = json.dumps(data, separators=(',', ':'), ensure_ascii=ensure_ascii)
    json_str += '\n'

    # Write to file
    getattr(file, 'atomic_write' if atomic else 'write')(file_path, json_str, encoding=encoding)



import json, re, urllib.request as http
from pathlib import Path
from typing import Union, List, Optional

api = json.loads((Path(__file__).parent / 'data/package_data.json').read_text())

def get_min_py(pkg_names: Union[str, List[str]]) -> Union[Optional[str], List[Optional[str]]]:
    if isinstance(pkg_names, str) : pkg_names = [pkg_names]
    results: List[Optional[str]] = []

    for pkg_name in pkg_names: # get min py
        try:
            req = http.Request(f'https://pypi.org/pypi/{pkg_name}/json')
            req.add_header('User-Agent', f"{api['name']}/{api['version']}")
            resp = http.urlopen(req, timeout=5)
            pkg_info = json.loads(resp.read())['info']

            # Check `requires_python`
            requires_python = pkg_info.get('requires_python')
            if requires_python:
                ver_match = re.search(r'(>|>=|==|~=)\s*(\d+\.\d+(?:\.\d+)?)', requires_python)
                if ver_match:
                    op, version = ver_match.group(1), ver_match.group(2)
                    if op == '>': # return minor-bumped
                        major, minor = version.split('.')[:2]
                        results.append(f'{major}.{int(minor) + 1}')
                    else: # >=|==|~= no bump
                        if version.endswith('.0') : version = version[:-2]
                        results.append(version)
                    continue # to next pkg

            # Check classifiers
            classifiers, versions = pkg_info.get('classifiers', []), []
            for classifier in classifiers:
                if classifier.startswith('Programming Language :: Python ::'):
                    ver_match = re.search(r'(\d+(?:\.\d+)?)', classifier)
                    if ver_match : versions.append(ver_match.group())
            if versions: # append lowest
                decimal_vers = [ver for ver in versions if '.' in ver]
                if decimal_vers:
                    decimal_vers.sort(key=lambda ver: [int(x) for x in ver.split('.')])
                    results.append(decimal_vers[0])
                else:
                    results.append(min(versions, key=int))
                continue # to next pkg
            else : results.append(None)

        except Exception : results.append(None)

    return results[0] if len(pkg_names) == 1 else results



import sys
from types import SimpleNamespace as sn
from typing import Optional

from . import log, string, url

controls = sn(
    help=sn(
        args=['-h', '--help'], action='help'),
    version=sn(
        args=['-v', '--version'], action='store_true', exit=True, handler=lambda cli: log.version(cli)),
    docs=sn(
        args=['--docs'], action='store_true', exit=True, handler=lambda cli: url.open(cli.urls.docs)),
    debug=sn(
        args=['-V', '--debug'], nargs='?', const=True, metavar='TARGET_KEY')
)

def get_canonical_key(key: str) -> Optional[str]:
    if key.startswith('-'): # convert CLI arg to full key name
        for ctrl_key, ctrl in vars(controls).items():
            if key in getattr(ctrl, 'args', []):
                key = ctrl_key
                break
    legacy_key = key if key.startswith('legacy_') else f'legacy_{key}'
    legacy_ctrl = getattr(controls, legacy_key, None)
    stripped_key = string.removeprefix(key, 'legacy_')
    return legacy_ctrl.replaced_by if legacy_ctrl and hasattr(legacy_ctrl, 'replaced_by') \
      else stripped_key if hasattr(controls, stripped_key) \
      else None

def is_neg_key(key: str) -> bool:
    import re
    return bool(re.match(r'^(?:no|disable|exclude)_', string.removeprefix(key, 'legacy_')))

def load(cli: sn) -> None:
    import argparse
    from . import data

    cli.config = sn()

    # Assign help tips from cli.msgs
    for ctrl_key, ctrl in vars(controls).items():
        if ctrl_key.startswith('legacy_') : continue
        if not hasattr(ctrl, 'help') : ctrl.help = getattr(cli.msgs, f'help_{ctrl_key.upper()}')

    # Parse CLI args
    argp = argparse.ArgumentParser(description=cli.description, add_help=False)
    argp.usage = cli.cmd_format
    valid_argparse_kwargs = {
        'action', 'choices', 'const', 'default', 'dest', 'help', 'metavar', 'nargs', 'required', 'type', 'version'}
    for ctrl_key, ctrl in vars(controls).items(): # add args to argp
        kwargs = ctrl.__dict__.copy()
        args = kwargs.pop('args')
        argparse_kwargs = { key:val for key,val in kwargs.items() if key in valid_argparse_kwargs }
        if ctrl_key.startswith('legacy_'): # copy canonical attrs first
            canonical_key = get_canonical_key(ctrl_key)
            if canonical_key: # adjust argparse_kwargs
                canonical_ctrl = getattr(controls, canonical_key)
                argparse_kwargs.update({
                    key:val for key,val in canonical_ctrl.__dict__.items() if key in valid_argparse_kwargs })
                argparse_kwargs['dest'] = canonical_key
                if is_neg_key(ctrl_key) != is_neg_key(canonical_key):
                    argparse_kwargs['action'] = 'store_false' if argparse_kwargs['action'] == 'store_true' \
                                           else 'store_true'
                for arg in args:
                    if arg in sys.argv[1:]:
                        log.warn_legacy_option(cli, arg, source='cli')
                        break
        argp.add_argument(*args, **argparse_kwargs)
    parsed_args, unknown_args = argp.parse_known_args()
    exempt_flags = [] # exempt valid dash-less args from validation
    exempt_flags.extend(arg.lstrip('-') for ctrl_key, ctrl in vars(controls).items()
                        if getattr(ctrl, 'subcmd', False)
                        for arg in ctrl.args if len(arg) > 2) # skip short flags
    if unknown_args:
        unknown_flags = [arg for arg in unknown_args if arg.startswith('-')]
        if unknown_flags:
            log.cmd_docs_url_exit(cli, f"{cli.msgs.err_UNRECOGNIZED_ARGS}: {' '.join(unknown_flags)}", cmd='help')
    for ctrl_key, ctrl in vars(controls).items(): # process subcmds
        if getattr(ctrl, 'subcmd', False) \
                and next(arg for arg in ctrl.args if arg.startswith('--'))[2:] in sys.argv[1:]:
            setattr(parsed_args, ctrl_key, True)
    applied_args = []
    for arg in sys.argv[1:]:
        if not arg.startswith('-') : continue
        base_arg = arg.split('=')[0]
        for ctrl_key, ctrl in vars(controls).items():
            if base_arg in getattr(ctrl, 'args', []):
                dest = get_canonical_key(ctrl_key) or ctrl_key if ctrl_key.startswith('legacy_') else ctrl_key
                parsed_val = getattr(parsed_args, dest, None)
                if parsed_val is not None:
                    setattr(cli.config, dest, parsed_val)
                    applied_args.append(arg)
                break
    log.debug(f'Args parsed! {log.colors.bg}{len(applied_args)} args applied {applied_args}', cli)

    # Apply parsers/default_vals
    for ctrl_key, ctrl in vars(controls).items():
        if not hasattr(cli.config, ctrl_key):
            setattr(cli.config, ctrl_key, ctrl.default_val if hasattr(ctrl, 'default_val') else None)
        config_val = getattr(cli.config, ctrl_key)
        if getattr(ctrl, 'parser', '') == 'csv':
            setattr(cli.config, ctrl_key, data.csv.parse(config_val))
    log.debug('All cli.config vals set!', cli)



from typing import List, Optional, Tuple

def get(url: str, timeout: int = 5, encoding: str = 'utf-8') -> str:
    from urllib.error import URLError
    from urllib.request import urlopen
    url = validate(url)
    try:
        with urlopen(url, timeout=timeout) as resp:
            return resp.read().decode(encoding)
    except URLError as err:
        raise RuntimeError(f'Failed to fetch from {url}: {err}')

def open(url: str) -> None:
    import webbrowser
    url = validate(url)
    try : webbrowser.open(url)
    except Exception as err:
        raise RuntimeError(f'Failed to open {url} in browser: {err}')

def validate(url: str, allowed_schemes: Tuple[str, ...] = ('http', 'https'),
                       allowed_domains: Optional[List[str]] = None) -> str:
    from urllib.parse import urlparse
    parsed_url = urlparse(url)

    if parsed_url.scheme not in allowed_schemes:
        raise ValueError(f'URL scheme {parsed_url.scheme!r} not allowed. Allowed: {allowed_schemes}')

    if allowed_domains:
        input_domain = parsed_url.netloc.lower()
        if not any(input_domain.endswith(allowed_domain.lower()) for allowed_domain in allowed_domains):
            raise ValueError(f'URL domain {input_domain!r} not allowed. Allowed: {allowed_domains}')

    return url



import sys

from ..api import get_min_py
from .lib import init, log, settings

def main():
    if len(sys.argv) == 1 : sys.argv.append('--help')
    cli = init.cli()

    # Process early-exit args (e.g. --help, --version)
    for ctrl_name, ctrl in vars(settings.controls).items():
        if getattr(ctrl, 'exit', False) and getattr(cli.config, ctrl_name, False):
            if hasattr(ctrl, 'handler') : ctrl.handler(cli)
            sys.exit(0)

    # Process pkgs
    pkgs = []
    for arg in sys.argv[1:]:
        if arg.startswith('-') : continue
        pkgs.extend([pkg.strip() for pkg in arg.split(',') if pkg.strip()])
    if pkgs:
        results = get_min_py(pkgs)
        log.package_vers(pkgs, results, cli)

if __name__ == '__main__' : main()



from types import SimpleNamespace as sn
from typing import Optional

def create_pkg_ver_url(cli: sn, version: Optional[str] = None) -> str:
    version = version or cli.version
    return f'{cli.urls.jsdelivr}@{cli.name}-{cli.version}/{cli.name}'

def create_commit_url(cli: sn, hash: str = 'latest') -> str:
    return f'{cli.urls.jsdelivr}@{hash}/{cli.name}'



import re

def get_next_maj_ver(version: str) -> str: # e.g. '1.2.3' -> '2.0.0'
    major = re.match(r'^(\d+)\..*', version)
    if not major : raise ValueError(f'Invalid version string: {version!r}')
    return f'{ int(major.group(1)) +1 }.0.0'
